# -*- coding: utf-8 -*-
import re

from transkribus.xml import XmlElement

CSS_RULES_REGEX = re.compile(r"(\S+)\s*{([^}]+)}")
CSS_PROPERTIES_REGEX = re.compile(r"(\S+?):\s*(\S+?);")
CSS_ESCAPE_REGEX = re.compile(r"\\u([0-9a-f]{4})", re.IGNORECASE)


def unescape_css_string(value):
    """
    Remove Transkribus' "CSS escapes": \\u followed by 4 characters to denote Unicode characters.
    https://gitlab.com/readcoop/transkribus/TranskribusCore/-/blob/f2deecaa1233126070e64ccc4e1eaa48d905fd11/src/main/java/eu/transkribus/core/model/beans/customtags/CssSyntaxTag.java#L29-49
    """
    return CSS_ESCAPE_REGEX.sub(lambda match: chr(int(match.group(1), 16)), value)


class Tag(object):
    def __init__(self, name, value):
        self.name = name
        if isinstance(value, dict):
            self.__dict__.update(value)
        for match in CSS_PROPERTIES_REGEX.finditer(value):
            prop_name, prop_value = match.groups()
            # Some special characters are escaped using \uXXXX; this unescapes them
            prop_value = unescape_css_string(prop_value)
            setattr(self, prop_name, prop_value)

    def as_dict(self):
        return self.__dict__

    @classmethod
    def build(cls, text):
        if not text:
            return []
        tags = []
        for match in CSS_RULES_REGEX.finditer(text):
            tags.append(cls(*match.groups()))
        return tags


class PageXmlElement(XmlElement):
    def __init__(self, path):
        super().__init__(path)
        nsmap = set(
            [
                self.element.nsmap.get(idx)
                for idx in self.element.nsmap
                if self.element.nsmap.get(idx).startswith(
                    "http://schema.primaresearch.org/PAGE/gts/pagecontent/"
                )
            ]
        )
        if len(nsmap) == 1:
            self.namespaces = {"page": list(nsmap)[0]}
        elif len(nsmap) > 1:
            raise Exception("can't manage different versions at the same time")
        else:
            raise Exception("There are no URLs, the PAGE namespace is missing")
        self._data = self.parse()


class Region(PageXmlElement):
    optional = ("custom",)

    def get_points(self, path):
        text = self._find(path)
        if text is None:
            # if Coords doesn't have a points attribute we check if it
            # has Point elements as children, to support versions of PAGE
            # prior to 2013-07-15

            points = []
            list_x = self._find("page:Coords/page:Point/@x", many=True)
            list_y = self._find("page:Coords/page:Point/@y", many=True)
            if list_x is None or list_y is None:
                return
            assert len(list_x) == len(
                list_y
            ), "the list of x and y coordinates do not have the same lengths"
            for x, y in zip(list_x, list_y):
                points.append((int(x), int(y)))
            return points
        else:
            points = []
            for coords in filter(None, map(str.strip, text.split(" "))):
                x, y = coords.split(",", 1)
                points.append((int(x), int(y)))
            return points

    def parse(self):
        return {
            "id": self.get_text("@id"),
            "points": self.get_points("page:Coords/@points"),
            "tags": Tag.build(self.get_text("@custom")),
        }


class TextLine(Region):
    optional = Region.optional + ("confidence",)

    def parse(self):
        data = super().parse()
        data.update(
            text=self.get_text("page:TextEquiv/page:Unicode"),
            baseline=self.get_points("page:Baseline/@points"),
            confidence=self.get_float("page:TextEquiv/@conf"),
        )
        return data


class TextRegion(Region):
    optional = Region.optional + ("type", "confidence")

    def parse(self):
        data = super().parse()
        data.update(
            type=self.get_text("@type"),
            text=self.get_text("page:TextEquiv/page:Unicode"),
            confidence=self.get_float("page:TextEquiv/@conf"),
            lines=self.get_instance(TextLine, "page:TextLine", many=True),
        )
        return data


class PageElement(PageXmlElement):
    optional = ("type",)

    def get_ordering(self):
        return [
            ref.get("regionRef")
            for ref in sorted(
                self._find(
                    "page:ReadingOrder/page:OrderedGroup/page:RegionRefIndexed",
                    many=True,
                ),
                key=lambda elt: int(elt.get("index")),
            )
        ]

    def get_relations(self):
        relations = []
        for relation in self._find(
            'page:Relations/page:Relation[@type="link"]', many=True
        ):
            refs = relation.findall("page:RegionRef", namespaces=self.namespaces)
            assert (
                len(refs) >= 2
            ), "Expected at least two regions in relation, got {}".format(len(refs))
            relations.append([ref.get("regionRef") for ref in refs])
        return relations

    def parse(self):
        return {
            "image_name": self.get_text("@imageFilename"),
            "image_width": self.get_text("@imageWidth"),
            "image_height": self.get_text("@imageHeight"),
            "type": self.get_text("@type"),
            "text_regions": self.get_instance(TextRegion, "page:TextRegion", many=True),
            "unknown_regions": self.get_instance(
                Region, "page:UnknownRegion", many=True
            ),
            # TODO? MusicRegion, NoiseRegion, TableRegion, SeparatorRegion
            "ordering": self.get_ordering(),
            "relations": self.get_relations(),
        }

    def sort_regions(self, regions):
        if len(regions) <= 1 or not self.ordering:
            return regions
        return sorted(
            regions,
            key=lambda region: self.ordering.index(region.id)
            if region.id in self.ordering
            else 0,
        )


class PageXmlMetadata(PageXmlElement):
    optional = ("comments",)

    def parse(self):
        return {
            "creator": self.get_text("page:Creator"),
            "created": self.get_text("page:Created"),
            "last_change": self.get_text("page:LastChange"),
            "comments": self.get_text("page:Comments"),
        }


class PageXmlPage(PageXmlElement):
    def parse(self):
        return {
            "metadata": self.get_instance(PageXmlMetadata, "page:Metadata"),
            "page": self.get_instance(PageElement, "page:Page"),
        }



# -*- coding: utf-8 -*-
import logging
import os

import requests

from transkribus.utils import try_request

logger = logging.getLogger(__name__)


def options_from_env():
    options = {}
    if "TRANSKRIBUS_API_URL" in os.environ:
        options["base_url"] = os.environ["TRANSKRIBUS_API_URL"]
    if "TRANSKRIBUS_EMAIL" in os.environ:
        options["email"] = os.environ["TRANSKRIBUS_EMAIL"]
    if "TRANSKRIBUS_PASSWORD" in os.environ:
        options["password"] = os.environ["TRANSKRIBUS_PASSWORD"]
    return options


class TranskribusAPI(object):
    def __init__(
        self,
        base_url="https://transkribus.eu/TrpServer/rest",
        email=None,
        password=None,
    ):
        self.base_url = base_url
        assert not self.base_url.endswith("/")
        assert self.base_url.startswith("https://")
        self.session = requests.Session()
        if email and password:
            self.login(email, password)

    def request(
        self,
        method,
        endpoint,
        params=None,
        accept="json",
        send="html",
        retries=3,
        cooldown=2.0,
        **data,
    ):
        assert not endpoint.startswith("/")
        assert accept in ("html", "json")
        assert send in ("html", "json")
        url = "{}/{}".format(self.base_url, endpoint)
        headers = {
            "User-Agent": "Arkindex/1.0",
        }

        if accept == "json":
            headers["Accept"] = "application/json"
        elif accept == "html":
            headers["Accept"] = "text/html, text/plain"
        else:
            raise Exception("Unsupported accept {}".format(accept))

        payload = {
            "headers": headers,
            "params": params,
            "data": send == "html" and data,
            "json": send == "json" and data,
        }

        resp = try_request(method, url, **payload)

        content_type = resp.headers.get("Content-Type")
        if content_type is not None and content_type.startswith("application/json"):
            return resp.json()

        return resp.content

    def get(self, *args, **kwargs):
        return self.request(self.session.get, *args, **kwargs)

    def post(self, *args, **kwargs):
        return self.request(self.session.post, *args, **kwargs)

    def login(self, email, password):
        user = self.post("auth/login", user=email, pw=password)
        logger.info("Logged as {userName} #{userId}".format(**user))
        return user

    def get_collection(self, collection_id):
        return self.get("collections/{}/metadata".format(collection_id))

    def create_collection(self, name):
        assert isinstance(name, str)
        col_id = self.post(
            "collections/createCollection", accept="html", params={"collName": name}
        )
        return col_id.decode("utf-8")

    def update_collection(self, collection_id, name=None, description=None):
        payload = self.get_collection(collection_id)
        if name is not None:
            assert isinstance(name, str)
            payload["colName"] = name
        if description is not None:
            assert isinstance(description, str)
            payload["description"] = description
        return self.post(
            "collections/{}/metadata".format(collection_id), send="json", **payload
        )

    def list_collections(self):
        return self.get("collections/list")

    def load_collection_documents(self, collection_id):
        return self.get("collections/{}/list".format(collection_id))

    def load_document(self, collection_id, document_id):
        return self.get("collections/{}/{}/fulldoc".format(collection_id, document_id))

    def list_user_collection(self, collection_id):
        return self.get("collections/{}/userlist".format(collection_id))

    def start_export(self, collection_id, **payload):
        return int(
            self.post(f"collections/{collection_id}/export", send="json", **payload)
        )

    def get_job(self, job_id):
        return self.get(f"jobs/{job_id}")



# -*- coding: utf-8 -*-
import logging
import time

from requests import HTTPError, RequestException

logger = logging.getLogger(__name__)


def try_request(method, *args, retries=3, cooldown=2.0, **kwargs):
    try:
        resp = method(*args, **kwargs)
        resp.raise_for_status()
        return resp
    except RequestException as e:
        if isinstance(e, HTTPError):
            msg = "HTTP error {}: {}".format(
                e.response.status_code,
                e.response.content.decode("utf-8"),
            )
        else:
            msg = str(e)

        if retries <= 1:
            raise Exception(msg)

        logger.warning(msg)
        logger.debug("Waiting for {:.2f} seconds before retrying".format(cooldown))
        time.sleep(cooldown)
        return try_request(
            method, *args, retries=retries - 1, cooldown=cooldown * 2, **kwargs
        )



# -*- coding: utf-8 -*-
import json
import logging
import os
import time

import requests
from humanize import naturalsize

from transkribus.utils import try_request

logger = logging.getLogger(__name__)

JOB_POLLING_SECONDS = 5


class Model(object):
    def __init__(self, data):
        # Allow instantiating using the ID directly
        if isinstance(data, int):
            data = {self.KEY: data}

        assert isinstance(data, dict)
        self.data = data
        self.id = data[self.KEY]
        self.directory = None

    def __getitem__(self, key):
        if key in self.data:
            return self.data[key]

    def build_directory(self, base):
        # Check base folder
        base = os.path.realpath(base)
        assert os.path.isdir(base)
        assert os.access(base, os.W_OK | os.X_OK)

        # Init model directory
        self.directory = os.path.join(base, str(self.id))
        if not os.path.isdir(self.directory):
            os.makedirs(self.directory)

        return self.directory

    def write_json(self, name, payload):
        assert self.directory is not None
        with open(os.path.join(self.directory, "{}.json".format(name)), "w") as f:
            json.dump(payload, f, indent=4, sort_keys=True)

    def download_image(self, url, filename=None, retries=3, cooldown=2.0):
        resp = try_request(requests.get, url, stream=True)
        # TODO: check content type

        if not filename:
            return resp.content

        # Get extension from filename
        # and build new file
        assert self.directory is not None
        _, ext = os.path.splitext(filename)
        path = os.path.join(self.directory, "{}{}".format(self.id, ext.lower()))

        with open(path, "wb") as f:
            for chunk in resp:
                f.write(chunk)

        logger.info("Downloaded image {}".format(path))
        return path

    def download_text(self, url, filename=None):
        resp = try_request(requests.get, url)
        # TODO: check encoding for binary

        if not filename:
            return resp.content

        assert self.directory is not None
        path = os.path.join(self.directory, filename)

        with open(path, "wb") as f:
            f.write(resp.content)

        logger.info("Downloaded text {}".format(path))
        return path

    def download(self, api, output_dir):
        raise NotImplementedError


class Collection(Model):
    KEY = "collectionId"

    def get_documents(self, api):
        return [Document(self, d) for d in api.load_collection_documents(self.id)]

    def download(self, api, output_dir):
        self.build_directory(output_dir)
        logger.info("Syncing collection {} to {}".format(self.id, self.directory))

        for doc in self.get_documents(api):
            logger.info("Sync document {}".format(doc))
            doc.download(api, self.directory)

    def export(self, api, export_params={}):
        return Job(api.start_export(self.id, **export_params))


class Document(Model):
    KEY = "docId"

    def __init__(self, collection, data):
        assert isinstance(collection, Collection)
        self.collection = collection
        super().__init__(data)
        self.base_data = self.data.copy()

    def __str__(self):
        return "{docId} - {title}".format(**self.data)

    def load(self, api):
        self.data.update(api.load_document(self.collection.id, self.id))

    def get_pages(self, api):
        if "pageList" not in self.data:
            self.load(api)
        return [Page(self, p) for p in self["pageList"]["pages"]]

    def download(self, api, output_dir):
        self.build_directory(output_dir)

        for page in self.get_pages(api):
            logger.info("Sync page {}".format(page))
            page.download(api, self.directory)

        # Dump payloads
        self.write_json("base", self.base_data)
        self.write_json("full", self.data)


class Page(Model):
    KEY = "pageId"

    def __init__(self, document, data):
        assert isinstance(document, Document)
        self.document = document
        super().__init__(data)

    def __str__(self):
        return "{pageId} - n°{pageNr}".format(**self.data)

    def get_transcript(self):
        # All the transcripts have a timestamp.
        # The transcript with the largest timestamp is the latest version of the transcripts.
        transcripts = self.data["tsList"]["transcripts"]
        last_transcript = max(transcripts, key=lambda tr: tr["timestamp"])
        return Transcript(self, last_transcript)

    def download(self, api, output_dir):
        self.directory = output_dir

        # Download full size image
        self.download_image(self.data["url"], self.data["imgFileName"])

        # Download transcription
        self.get_transcript().download(api, self.directory)


class Transcript(Model):
    KEY = "tsId"

    def __init__(self, page, data):
        assert isinstance(page, Page)
        self.page = page
        super().__init__(data)

    def __str__(self):
        return "{tsId} - {status} - {nbOfLines} lines".format(**self.data)

    def download(self, api, output_dir):
        self.directory = output_dir
        self.download_text(self.data["url"], "{}.xml".format(self.id))


class Job(Model):
    KEY = "jobId"

    def __str__(self) -> str:
        return "{jobId} - {status} - {description}".format(**self.data)

    @property
    def completion(self) -> float:
        if self.data.get("progress") is None or self.data.get("totalWork") is None:
            return
        # Avoid dividing by zero
        if self.data["totalWork"] == 0:
            return 0.0
        return self.data["progress"] / self.data["totalWork"]

    def refresh(self, api) -> None:
        """
        Reload the job's data from the API.
        """
        self.data = api.get_job(self.id)

    def wait_for_result(self, api, timeout=1800) -> None:
        """
        Poll the API every few seconds until the job has a result URL or until a timeout.
        Raises an exception if the timeout is reached.
        """
        start_time = time.time()
        previous_description, previous_completion = (
            self.data.get("description"),
            self.completion,
        )

        while start_time + timeout > time.time() and not self.data.get("result"):
            time.sleep(JOB_POLLING_SECONDS)
            self.refresh(api)

            # Log status changes
            if (
                self.data.get("description") != previous_description
                or self.completion != previous_completion
            ):
                logger.info(
                    f'Job {self.id} ({int(self.completion * 100)}%): {self["description"]}'
                )
                previous_description, previous_completion = (
                    self.data.get("description"),
                    self.completion,
                )

        assert self.data.get(
            "result"
        ), f"Timed out while waiting for a result from job {self.id}"

    def download_result(self, destination) -> None:
        assert self.data.get("result"), "Job has no result URL"
        resp = try_request(requests.get, self.data["result"], stream=True)

        size = int(resp.headers["Content-Length"])
        logger.info(f"Downloading {naturalsize(size)}")

        downloaded, last_percent = 0, 0
        with open(destination, "wb") as f:
            # Save the file in 1MB chunks
            for chunk in resp.iter_content(chunk_size=1024**2):
                f.write(chunk)
                downloaded += len(chunk)

                # Log only when the percentage changes
                percent = int((downloaded / size) * 100)
                if percent != last_percent:
                    last_percent = percent
                    logger.info(
                        f"Downloaded {naturalsize(downloaded)} / {naturalsize(size)} ({percent}%)"
                    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from setuptools import find_packages, setup


def read_requirements(filename):
    return [req.strip() for req in open(filename)]


setup(
    name="transkribus-client",
    version=open("VERSION").read().strip(),
    author="Teklia <contact@teklia.com>",
    packages=find_packages(
        exclude=["*.tests", "*.tests.*", "tests.*", "tests"],
    ),
    package_data={
        "": ["README.md", "LICENSE"],
    },
    install_requires=read_requirements("requirements.txt"),
    python_requires=">=3.6",
    license="MIT",
    description="Unofficial API client for the Transkribus project",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    keywords="api client transkribus",
    url="https://gitlab.com/arkindex/transkribus",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering :: Image Recognition",
        "Topic :: Text Processing :: Indexing",
        "Topic :: Text Processing :: Linguistic",
    ],
)
